@@ -2027,10 +2027,6 @@ my $known = {
       70 => 'flags',
       1 => 'text_value',
     ],
-  ACDB_BLOCKREPRESENTATION_DATA => [
-    70 => 'flag',
-    340 => 'block',
-    ],
   ACMECOMMANDHISTORY => [
     ],
   ACMESCOPE => [
