@@ -44,55 +44,20 @@
 //void *memmem (const void *big, size_t big_len, const void *little,
 //              size_t little_len);
 
+
+#define CHAIN_SIZE 128
 int cur_hdl; // to avoid dupl. search
 
-struct _unknown_field
-{
-  int code;
-  const char *value;
-  unsigned char *bytes;
-  int num_bits;
-  Dwg_Bits type;
-  const char *name; // in dwg.spec
-  // number of occurrences of this type:value pair (max 1423)
-  unsigned short num;
-  int pos[5]; // 5x found bit offset in dxf->bytes or -1 if not found
-  // many typical values are 5x found (handle 0, BL 2)
-};
-static struct _unknown_dxf
-{
-  const char *name;
-  const char *dxf;
-  const unsigned int handle;
-  const char *bytes;
-  const int is_entity;
-  const int num_bits; // size of dumped unknown_bits TF
-  const int commonsize;
-  const int hdloff;
-  const int strsize;
-  const int hdlsize;
-  const int bitsize;
-  const struct _unknown_field *fields;
-} unknown_dxf[] = {
-// see log_unknown_dxf.pl
-#include "alldxf_0.inc"
+#include "unknown.h"
+
+static struct _unknown_dxf unknown_dxf[] = {
+  // see log_unknown_dxf.pl
+  #include "alldxf_0.inc"
   { NULL, NULL, 0, "", 0, 0, 0, 0, 0, 0, 0, NULL }
 };
 
 #include "alldxf_1.inc"
 
-// dynamic helper companion
-struct _dxf
-{
-  unsigned char *found;    // coverage per bit for found 1
-  unsigned char *possible; // coverage for mult. finds >1
-  int num_bits;            // copy of unknown_dxf.num_bits
-  int num_filled;
-  int num_empty;
-  int num_possible;
-  double percentage;
-};
-
 /* not needed for the solver, only to check against afterwards */
 #if 0
 static struct _unknown {
@@ -110,7 +75,7 @@ static struct _unknown {
   const int bitsize;
 } unknowns[] =
   {
-   { "ACAD_EVALUATION_GRAPH", "example_2000.log", "40501406481013fffffffcffffffff3fffffffcffffffff980c0c80b8bee", "test/test-data/example_2000.dxf", 0x2E3, 0, 234, 60, -60, 0, 268 },
+   { "ACAD_EVALUATION_GRAPH", "example_2000.log", 0x2E3, "40501406481013fffffffcffffffff3fffffffcffffffff980c0c80b8bee", "test/test-data/example_2000.dxf", 0, 234, 60, -60, 0, 268 },
     /* the following types:
       5 ACDBASSOCGEOMDEPENDENCY
       3 ACDBASSOCNETWORK
@@ -135,8 +100,8 @@ static struct _bd
   const char *value;
   const char *bin;
 } bd[] = {
-// see bd-unknown.pl
-#include "bd-unknown.inc"
+  // see bd-unknown.pl
+  #include "bd-unknown.inc"
   { NULL, NULL }
 };
 
@@ -150,14 +115,17 @@ bits_TV (Bit_Chain *restrict dat, struct _unknown_field *restrict g)
 static void
 bits_TU (Bit_Chain *restrict dat, struct _unknown_field *restrict g)
 {
-  bit_write_TU (dat, (BITCODE_TU)g->value);
+  // the source is always utf8
+  BITCODE_TU wstr = bit_utf8_to_TU ((char*)g->value);
+  bit_write_TU (dat, wstr);
   g->type = BITS_TU;
+  free (wstr);
 }
 
 static void
 bits_TF (Bit_Chain *restrict dat, struct _unknown_field *restrict g, int len)
 {
-  bit_write_TF (dat, (char *)g->value, len);
+  bit_write_TF (dat, (BITCODE_TF)g->value, len);
   g->type = BITS_TF;
 }
 
@@ -175,7 +143,7 @@ bits_hexstring (Bit_Chain *restrict dat, struct _unknown_field *restrict g)
 {
   // convert hex to string
   int len = strlen (g->value) / 2;
-  char buf[1024];
+  unsigned char buf[1024];
   for (int i = 0; i < len; i++)
     {
       unsigned char *s = (unsigned char *)&g->value[i * 2];
@@ -234,7 +202,7 @@ bits_BD (Bit_Chain *restrict dat, struct _unknown_field *restrict g)
   // some more not found BD values
   for (struct _bd *b = &bd[0]; b->value; b++)
     {
-      if (!strcmp (g->value, b->value))
+      if (strEQ (g->value, b->value))
         {
           bit_write_bits (dat, b->bin);
           return;
@@ -327,7 +295,7 @@ bits_CMC (Bit_Chain *restrict dat, struct _unknown_field *restrict g)
             }
         }
     }
-  bit_write_CMC (dat, &color);
+  bit_write_CMC (dat, dat, &color);
   g->type = BITS_CMC;
 }
 
@@ -471,8 +439,8 @@ is_handle (int code)
 static void
 bits_try_handle (struct _unknown_field *g, int code, unsigned int objhandle)
 {
-  Bit_Chain dat = { NULL, 16, 0, 0, NULL, 0, 0 };
-  dat.chain = calloc (1, 16);
+  Bit_Chain dat = EMPTY_CHAIN (CHAIN_SIZE);
+  dat.chain = calloc (1, CHAIN_SIZE);
 
   bits_handle (&dat, g, code, objhandle);
 
@@ -494,10 +462,10 @@ bits_format (struct _unknown_field *g, const int version,
              struct _unknown_dxf *dxf)
 {
   int code = g->code;
-  Bit_Chain dat = { NULL, 16, 0, 0, NULL, 0, 0 };
+  Bit_Chain dat = EMPTY_CHAIN (CHAIN_SIZE);
   const int is16 = version >= 2007 ? 1 : 0;
 
-  dat.chain = calloc (1, 16);
+  dat.chain = calloc (1, CHAIN_SIZE);
   if (version)
     {
       char s[16];
@@ -524,7 +492,10 @@ bits_format (struct _unknown_field *g, const int version,
         bits_CMC (&dat, g);
     }
   else if (code < 80)
-    bits_BS (&dat, g);
+    if (code == 75 && strEQc (g->name, "ref.has_lastpt_ref"))
+      bits_B (&dat, g);
+    else
+      bits_BS (&dat, g);
   else if (80 <= code && code <= 99) // BL int32
     bits_BL (&dat, g);
   else if (code == 100)
@@ -578,7 +549,7 @@ bits_format (struct _unknown_field *g, const int version,
   else if (code <= 481)
     bits_handle (&dat, g, 5, 0);
   else if (code == 999)
-    return;
+    ;
   else if (1000 <= code && code <= 1009)
     bits_string (&dat, g);
   else if (1010 <= code && code <= 1049)
@@ -600,13 +571,14 @@ bits_format (struct _unknown_field *g, const int version,
   else
     {
       free (dat.chain);
+      dat.chain = NULL;
     }
 }
 
 // check how many of the given fields (code=value pairs) exist in the DXF
 #if 0
 static int
-num_dxf(const struct _unknown_field *g, const struct _unknown_dxf *dxf) {
+num_dxf (const struct _unknown_field *g, const struct _unknown_dxf *dxf) {
   int num_dxf = 0;
   struct _unknown_field *f = (struct _unknown_field *)dxf->fields;
   while (f->value) {
@@ -796,6 +768,57 @@ cquote (char *restrict dest, const char *restrict src)
   return d;
 }
 
+static void
+pi_filename (char *restrict pi_fn, uint16_t i)
+{
+  if (i)
+    {
+      char tmp[36];
+      snprintf (tmp, 35, ".%hu", i);
+      tmp[35] = '\0';
+      strcat (pi_fn, tmp);
+    }
+  strcat (pi_fn, ".pi");
+}
+
+static int
+open_pi (FILE *pi, char *restrict class, char *pi_fn)
+{
+  if (!pi)
+    {
+      fprintf (stderr, "Failed to write %s\n", pi_fn);
+      return 1;
+    }
+  fprintf (pi,
+           "import unknown.\n\n"
+           "/* %s field packing problem.\n"
+           "   examples/unknown generated example, needs picat-lang.org.\n"
+           "   Usage: picat [-g go2] %s\n"
+           "*/\n",
+           class, pi_fn);
+  return 0;
+}
+
+static void
+close_pi (FILE *pi, long class_filled, long class_size, int k)
+{
+  int i;
+  // class_summary
+  fprintf (pi, "\n%% summary: %ld/%ld=%.1f%%\n\n", class_filled, class_size,
+           100.0 * class_filled / class_size);
+
+  for (i = 0; i < k; i++)
+    {
+      fprintf (pi, "go%d ?=> def(%d,Data).\n", i, i);
+    }
+  fprintf (pi, "\nmain => go0%s\n", k == 1 ? "." : ",");
+  for (i = 1; i < k; i++)
+    {
+      fprintf (pi, "        go%d%s\n", i, i == k - 1 ? "." : ",");
+    }
+  fclose (pi);
+}
+
 int
 main (int argc, char *argv[])
 {
@@ -807,7 +830,7 @@ main (int argc, char *argv[])
   char *classes[MAX_CLASSES]; // create files per classes
   struct _dxf *dxf = calloc (sizeof (unknown_dxf) / sizeof (unknown_dxf[0]),
                              sizeof (struct _dxf));
-// clang-format off
+  // clang-format off
   #include "alldxf_2.inc"
   // clang-format on
 
@@ -846,9 +869,11 @@ main (int argc, char *argv[])
     {
       FILE *pi;
       char pi_fn[256];
+      char base_pi_fn[256];
       int k = 0;
       long class_filled = 0, class_size = 0;
       char *dn;
+      uint16_t i_pi = 0;
 
       class = classes[ic];
       // dirname should be examples/
@@ -863,21 +888,14 @@ main (int argc, char *argv[])
         {
           strcpy (pi_fn, "examples/");
         }
+      else
+          strcpy (pi_fn, "");
       strcat (pi_fn, class);
-      strcat (pi_fn, ".pi");
+      strcpy (base_pi_fn, pi_fn);
+      pi_filename (pi_fn, i_pi);
       pi = fopen (pi_fn, "w");
-      if (!pi)
-        {
-          fprintf (stderr, "Failed to write %s\n", pi_fn);
+      if (open_pi (pi, class, pi_fn))
           continue;
-        }
-      fprintf (pi,
-               "import unknown.\n\n"
-               "/* %s field packing problem.\n"
-               "   examples/unknown generated example, needs picat-lang.org.\n"
-               "   Usage: picat [-g go2] %s\n"
-               "*/\n",
-               class, pi_fn);
       for (i = 0; unknown_dxf[i].name; i++)
         {
           int num_fields;
@@ -1008,40 +1026,34 @@ main (int argc, char *argv[])
                     }
                   if (g[j].type == BITS_ENC)
                     {
-                      Bit_Chain dat = { NULL, 16, 0, 0, NULL, 0, 0 };
-                      dat.chain = calloc (1, 16);
+                      Bit_Chain dat = EMPTY_CHAIN (CHAIN_SIZE);
+                      dat.chain = calloc (1, CHAIN_SIZE);
 
                       bits_CMC (&dat, &g[j]);
                       g[j].bytes = dat.chain;
                       g[j].num_bits = (dat.byte * 8) + dat.bit;
                       num_found = search_bits (j, &g[j], &unknown_dxf[i],
                                                &dxf[i], offset);
-                      free (dat.chain);
                       if (num_found)
-                        {
-                          goto FOUND;
-                        }
+                        goto FOUND;
                     }
                   if (g[j].type == BITS_BS && strlen (g[j].value) < 3)
                     {
-                      Bit_Chain dat = { NULL, 16, 0, 0, NULL, 0, 0 };
-                      dat.chain = calloc (1, 16);
+                      Bit_Chain dat = EMPTY_CHAIN (CHAIN_SIZE);
+                      dat.chain = calloc (1, CHAIN_SIZE);
 
                       bits_RC (&dat, &g[j]);
                       g[j].bytes = dat.chain;
                       g[j].num_bits = (dat.byte * 8) + dat.bit;
                       num_found = search_bits (j, &g[j], &unknown_dxf[i],
                                                &dxf[i], offset);
-                      free (dat.chain);
                       if (num_found)
-                        {
-                          goto FOUND;
-                        }
+                        goto FOUND;
                     }
                   if (g[j].type == BITS_BS)
                     {
-                      Bit_Chain dat = { NULL, 16, 0, 0, NULL, 0, 0 };
-                      dat.chain = calloc (1, 16);
+                      Bit_Chain dat = EMPTY_CHAIN (CHAIN_SIZE);
+                      dat.chain = calloc (1, CHAIN_SIZE);
 
                       bits_BL (&dat, &g[j]);
                       g[j].bytes = dat.chain;
@@ -1049,10 +1061,7 @@ main (int argc, char *argv[])
                       num_found = search_bits (j, &g[j], &unknown_dxf[i],
                                                &dxf[i], offset);
                       if (num_found)
-                        {
-                          free (dat.chain);
-                          goto FOUND;
-                        }
+                        goto FOUND;
 
                       bit_set_position (&dat, 0);
                       bits_RS (&dat, &g[j]);
@@ -1060,17 +1069,14 @@ main (int argc, char *argv[])
                       g[j].num_bits = (dat.byte * 8) + dat.bit;
                       num_found = search_bits (j, &g[j], &unknown_dxf[i],
                                                &dxf[i], offset);
-                      free (dat.chain);
                       if (num_found)
-                        {
-                          goto FOUND;
-                        }
+                        goto FOUND;
                     }
                   if ((g[j].type == BITS_BL || g[j].type == BITS_BLd)
                       && strlen (g[j].value) <= 5)
                     {
-                      Bit_Chain dat = { NULL, 16, 0, 0, NULL, 0, 0 };
-                      dat.chain = calloc (1, 16);
+                      Bit_Chain dat = EMPTY_CHAIN (CHAIN_SIZE);
+                      dat.chain = calloc (1, CHAIN_SIZE);
 
                       bits_BS (&dat, &g[j]);
                       g[j].bytes = dat.chain;
@@ -1078,10 +1084,7 @@ main (int argc, char *argv[])
                       num_found = search_bits (j, &g[j], &unknown_dxf[i],
                                                &dxf[i], offset);
                       if (num_found)
-                        {
-                          free (dat.chain);
-                          goto FOUND;
-                        }
+                        goto FOUND;
 
                       bit_set_position (&dat, 0);
                       bits_RL (&dat, &g[j]);
@@ -1089,43 +1092,34 @@ main (int argc, char *argv[])
                       g[j].num_bits = (dat.byte * 8) + dat.bit;
                       num_found = search_bits (j, &g[j], &unknown_dxf[i],
                                                &dxf[i], offset);
-                      free (dat.chain);
                       if (num_found)
-                        {
-                          goto FOUND;
-                        }
+                        goto FOUND;
                     }
                   if (g[j].type == BITS_RC)
                     {
-                      Bit_Chain dat = { NULL, 16, 0, 0, NULL, 0, 0 };
-                      dat.chain = calloc (1, 16);
+                      Bit_Chain dat = EMPTY_CHAIN (CHAIN_SIZE);
+                      dat.chain = calloc (1, CHAIN_SIZE);
 
                       bits_BS (&dat, &g[j]);
                       g[j].bytes = dat.chain;
                       g[j].num_bits = (dat.byte * 8) + dat.bit;
                       num_found = search_bits (j, &g[j], &unknown_dxf[i],
                                                &dxf[i], offset);
-                      free (dat.chain);
                       if (num_found)
-                        {
-                          goto FOUND;
-                        }
+                        goto FOUND;
                     }
                   if (g[j].type == BITS_BD)
                     {
-                      Bit_Chain dat = { NULL, 16, 0, 0, NULL, 0, 0 };
-                      dat.chain = calloc (1, 16);
+                      Bit_Chain dat = EMPTY_CHAIN (CHAIN_SIZE);
+                      dat.chain = calloc (1, CHAIN_SIZE);
 
                       bits_RD (&dat, &g[j]);
                       g[j].bytes = dat.chain;
                       g[j].num_bits = (dat.byte * 8) + dat.bit;
                       num_found = search_bits (j, &g[j], &unknown_dxf[i],
                                                &dxf[i], offset);
-                      free (dat.chain);
                       if (num_found)
-                        {
-                          goto FOUND;
-                        }
+                        goto FOUND;
                     }
                   // relaxed BD search, less mantissa precision.
                   // e.g. 49:"0.0008202099737533" (66 bits of type BD)
@@ -1133,8 +1127,8 @@ main (int argc, char *argv[])
                   if (g[j].type == BITS_RD && strlen (g[j].value) >= 3)
                     {
                       double d;
-                      Bit_Chain dat = { NULL, 16, 0, 0, NULL, 0, 0 };
-                      dat.chain = calloc (1, 16);
+                      Bit_Chain dat = EMPTY_CHAIN (CHAIN_SIZE);
+                      dat.chain = calloc (1, CHAIN_SIZE);
 
                       bits_BD (&dat, &g[j]); // g.value -> dat
                       g[j].bytes = dat.chain;
@@ -1142,10 +1136,7 @@ main (int argc, char *argv[])
                       if (dat.byte == 8)
                         g[j].num_bits = 58; // from 66
                       else
-                        {
-                          free (dat.chain);
-                          goto FOUND;
-                        }
+                        goto FOUND;
                       // print rounded found value and show bit diff
                       printf (
                           "  imprecise BD search, 42bit mantissa precision\n");
@@ -1155,6 +1146,7 @@ main (int argc, char *argv[])
                         {
                           free (dat.chain);
                           dat.chain = (unsigned char *)unknown_dxf[i].bytes;
+                          g[j].bytes = dat.chain;
                           dat.size = unknown_dxf[i].num_bits / 8;
                           bit_set_position (&dat, g[j].pos[0]);
                           d = bit_read_BD (&dat);
@@ -1177,11 +1169,11 @@ main (int argc, char *argv[])
                           // precision\n");
                           num_found = search_bits (j, &g[j], &unknown_dxf[i],
                                                    &dxf[i], offset);
-                          free (dat.chain);
                           if (num_found)
                             {
-                              dat.chain
-                                  = (unsigned char *)unknown_dxf[i].bytes;
+                              free (dat.chain);
+                              dat.chain = (unsigned char *)unknown_dxf[i].bytes;
+                              g[j].bytes = dat.chain;
                               dat.size = unknown_dxf[i].num_bits / 8;
                               bit_set_position (&dat, g[j].pos[0]);
                               d = bit_read_BD (&dat);
@@ -1204,8 +1196,8 @@ main (int argc, char *argv[])
                   if (g[j].type == BITS_RD && strlen (g[j].value) >= 3)
                     {
                       double d;
-                      Bit_Chain dat = { NULL, 16, 0, 0, NULL, 0, 0 };
-                      dat.chain = calloc (1, 16);
+                      Bit_Chain dat = EMPTY_CHAIN (CHAIN_SIZE);
+                      dat.chain = calloc (1, CHAIN_SIZE);
 
                       bits_RD (&dat, &g[j]); // g.value -> dat
                       g[j].bytes = dat.chain;
@@ -1213,10 +1205,7 @@ main (int argc, char *argv[])
                       if (dat.byte == 8)
                         g[j].num_bits = 56; // from 64
                       else
-                        {
-                          free (dat.chain);
-                          goto FOUND;
-                        }
+                        goto FOUND;
                       // print rounded found value and show bit diff
                       printf (
                           "  imprecise RD search, 42bit mantissa precision\n");
@@ -1249,11 +1238,10 @@ main (int argc, char *argv[])
                           // precision\n");
                           num_found = search_bits (j, &g[j], &unknown_dxf[i],
                                                    &dxf[i], offset);
-                          free (dat.chain);
                           if (num_found)
                             {
-                              dat.chain
-                                  = (unsigned char *)unknown_dxf[i].bytes;
+                              free (dat.chain);
+                              dat.chain = (unsigned char *)unknown_dxf[i].bytes;
                               dat.size = unknown_dxf[i].num_bits / 8;
                               bit_set_position (&dat, g[j].pos[0]);
                               d = bit_read_BD (&dat);
@@ -1278,9 +1266,9 @@ main (int argc, char *argv[])
                   // the length usually includes the final \0
                   if (g[j].type == BITS_TV || g[j].type == BITS_TU)
                     {
-                      Bit_Chain dat = { NULL, 16, 0, 0, NULL, 0, 0 };
+                      Bit_Chain dat = EMPTY_CHAIN (CHAIN_SIZE);
                       int len = strlen (g[j].value);
-                      dat.chain = calloc (1, 16);
+                      dat.chain = calloc (1, CHAIN_SIZE);
 
 #if 0
                       /* TU/TF cannot be mixed */
@@ -1293,28 +1281,22 @@ main (int argc, char *argv[])
                           num_found = search_bits (j, &g[j], &unknown_dxf[i],
                                                    &dxf[i], offset);
                           if (num_found)
-                            {
-                              free (dat.chain);
-                              goto FOUND;
-                            }
+                            goto FOUND;
                         }
 #endif
                       if (len)
                         {
                           bit_set_position (&dat, 0);
-                          bits_TF (&dat, &g[j],
-                                   len); // search without the final \0
+                          // search without the final \0
+                          bits_TF (&dat, &g[j], len);
 
                           g[j].bytes = dat.chain;
                           g[j].num_bits = (dat.byte * 8) + dat.bit;
                           num_found = search_bits (j, &g[j], &unknown_dxf[i],
                                                    &dxf[i], offset);
-                          free (dat.chain);
                           if (num_found)
                             goto FOUND;
                         }
-                      else
-                        free (dat.chain);
                     }
 
                   if (!num_found || is_common_entity_data (g[j].code))
@@ -1530,6 +1512,10 @@ main (int argc, char *argv[])
                 &g[j])) { offset = g[j].pos[0]+1; goto SEARCH;
                 }
                 */
+              if (0) {
+                free (g[j].bytes);
+                g[j].bytes = NULL;
+              }
             }
           fprintf (pi, "  ],\n"
                        "  Data = [S,Fields,Class,Dxf,Version,Offsets],\n"
@@ -1580,26 +1566,24 @@ main (int argc, char *argv[])
 
           // TODO: try likely field combinations and print the top 3.
           // See unknown.pi
-          // there are various heuristics, like the handle stream at the end
+          // there are various heuristics, like the handle and string stream at the end.
+          // points BD's being neighbors, ...
 
           free (dxf[i].found);
           free (dxf[i].possible);
           k++;
+          if (k >= 50)
+            {
+              strcpy (pi_fn, base_pi_fn);
+              pi_filename (pi_fn, ++i_pi);
+              close_pi (pi, class_filled, class_size, k);
+              k = 0;
+              pi = fopen (pi_fn, "w");
+              if (open_pi (pi, class, pi_fn))
+                continue;
+            }
         }
-      // class_summary
-      fprintf (pi, "\n%% summary: %ld/%ld=%.1f%%\n\n", class_filled,
-               class_size, 100.0 * class_filled / class_size);
-
-      for (i = 0; i < k; i++)
-        {
-          fprintf (pi, "go%d ?=> def(%d,Data).\n", i, i);
-        }
-      fprintf (pi, "\nmain => go0%s\n", k == 1 ? "." : ",");
-      for (i = 1; i < k; i++)
-        {
-          fprintf (pi, "        go%d%s\n", i, i == k - 1 ? "." : ",");
-        }
-      fclose (pi);
+      close_pi (pi, class_filled, class_size, k);
     }
 
   printf ("summary: %ld/%ld=%.2f%%\n", sum_filled, sum_size,
