@@ -1,7 +1,7 @@
 /*****************************************************************************/
 /*  LibreDWG - free implementation of the DWG file format                    */
 /*                                                                           */
-/*  Copyright (C) 2009-2019 Free Software Foundation, Inc.                   */
+/*  Copyright (C) 2009-2020 Free Software Foundation, Inc.                   */
 /*  Copyright (C) 2010 Thien-Thi Nguyen                                      */
 /*                                                                           */
 /*  This library is free software, licensed under the terms of the GNU       */
@@ -73,8 +73,7 @@ help (void)
 #else
   printf ("  -v[0-9]     verbosity\n");
   printf ("  -O fmt      fmt: DXF, DXFB, JSON, GeoJSON\n");
-  printf (
-      "              Planned output formats:  YAML, XML/OGR, GPX, SVG, PS\n");
+  printf ("              Planned output formats:  YAML, XML/OGR, GPX, SVG, PS\n");
   printf ("  -o outfile  also defines the output fmt. Default: stdout\n");
   printf ("  -h          display this help and exit\n");
   printf ("  -i          output version information and exit\n"
@@ -94,6 +93,7 @@ main (int argc, char *argv[])
   const char *fmt = NULL;
   const char *outfile = NULL;
   int has_v = 0;
+  int force_free = 0;
   int c;
 #ifdef HAVE_GETOPT_LONG
   int option_index = 0;
@@ -101,6 +101,7 @@ main (int argc, char *argv[])
       = { { "verbose", 1, &opts, 1 }, // optional
           { "format", 1, 0, 'O' },    { "file", 1, 0, 'o' },
           { "help", 0, 0, 0 },        { "version", 0, 0, 0 },
+          { "force-free", 0, 0, 0 },
           { NULL, 0, NULL, 0 } };
 #endif
 
@@ -151,6 +152,8 @@ main (int argc, char *argv[])
             return opt_version ();
           if (!strcmp (long_options[option_index].name, "help"))
             return help ();
+          if (!strcmp (long_options[option_index].name, "force-free"))
+            force_free = 1;
           break;
 #else
         case 'i':
@@ -208,12 +211,6 @@ main (int argc, char *argv[])
     }
   i = optind;
 
-  if (optind == argc)
-    {
-      fprintf (stderr, "No input file specified");
-      return 1;
-    }
-  // REQUIRE_INPUT_FILE_ARG (optind);
   memset (&dwg, 0, sizeof (Dwg_Data));
   if (has_v || !fmt)
     dwg.opts = opts;
@@ -222,37 +219,30 @@ main (int argc, char *argv[])
     setenv ("LIBREDWG_TRACE", "1", 0);
 #endif
 
-  if (opts > 1)
-    fprintf (stderr, "Reading DWG file %s\n", argv[i]);
-  // TODO support stdin as in dwgwrite
-  error = dwg_read_file (argv[i], &dwg);
-  if (!fmt)
+  if (optind != argc)
     {
-      if (error >= DWG_ERR_CRITICAL)
-        {
-          fprintf (stderr, "ERROR 0x%x\n", error);
-          if (error && opts > 2)
-            dwg_errstrings (error);
-        }
-      else
-        {
-          if (opts > 1)
-            {
-              fprintf (stderr, "SUCCESS 0x%x\n", error);
-              if (error && opts > 2)
-                dwg_errstrings (error);
-            }
-          else
-            fprintf (stderr, "SUCCESS\n");
-        }
+      if (opts > 1)
+        fprintf (stderr, "Reading DWG file %s\n", argv[i]);
+      error = dwg_read_file (argv[i], &dwg);
     }
   else
+    {
+      if (opts > 1)
+        fprintf (stderr, "Reading DWG from stdin\n");
+      error = dwg_read_file ("-", &dwg); // i.e. from stdin
+    }
+
+  if (error >= DWG_ERR_CRITICAL)
+    goto done;
+
+  if (fmt)
     {
       Bit_Chain dat = { 0 };
       if (outfile)
         dat.fh = fopen (outfile, "w");
       else
         dat.fh = stdout;
+      fprintf (stderr, "\n");
       dat.version = dat.from_version = dwg.header.version;
       // TODO --as-rNNNN version? for now not.
       // we want the native dump, converters are separate.
@@ -286,37 +276,49 @@ main (int argc, char *argv[])
         fprintf (stderr, "Invalid output format '%s'\n", fmt);
 
       if (outfile)
-        {
-          fclose (dat.fh);
-          if (error >= DWG_ERR_CRITICAL)
-            {
-              fprintf (stderr, "ERROR 0x%x\n", error);
-              if (error && opts > 2)
-                dwg_errstrings (error);
-            }
-          else
-            {
-              if (opts > 1)
-                {
-                  fprintf (stderr, "SUCCESS 0x%x\n", error);
-                  if (error && opts > 2)
-                    dwg_errstrings (error);
-                }
-              else
-                fprintf (stderr, "SUCCESS\n");
-            }
-        }
+        fclose (dat.fh);
     }
+
+ done:
+#if defined __SANITIZE_ADDRESS__ || __has_feature(address_sanitizer)
+  {
+    char *asanenv = getenv("ASAN_OPTIONS");
+    if (!asanenv)
+      force_free = 1;
+    // detect_leaks is enabled by default. see if it's turned off
+    else if (strstr (asanenv, "detect_leaks=0") == NULL) /* not found */
+      force_free = 1;
+  }
+#endif
+
   // forget about valgrind. really huge DWG's need endlessly here.
   if ((dwg.header.version && dwg.num_objects < 1000)
-#if defined __SANITIZE_ADDRESS__ || __has_feature(address_sanitizer)
-      || 1
-#endif
+      || force_free
 #ifdef HAVE_VALGRIND_VALGRIND_H
       || (RUNNING_ON_VALGRIND)
 #endif
   )
-    dwg_free (&dwg);
+    {
+      dwg_free (&dwg);
+    }
+
+  if (error >= DWG_ERR_CRITICAL)
+    {
+      fprintf (stderr, "ERROR 0x%x\n", error);
+      if (error && opts > 2)
+        dwg_errstrings (error);
+    }
+  else
+    {
+      if (opts > 1)
+        {
+          fprintf (stderr, "SUCCESS 0x%x\n", error);
+          if (error && opts > 2)
+            dwg_errstrings (error);
+        }
+      else
+        fprintf (stderr, "SUCCESS\n");
+    }
 
   return error >= DWG_ERR_CRITICAL ? 1 : 0;
 }
